# Copyright 2023 The Boax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The acquisitions sub-package."""

from . import surrogates as surrogates
from .alias import expected_improvement as expected_improvement
from .alias import log_expected_improvement as log_expected_improvement
from .alias import (
  log_probability_of_improvement as log_probability_of_improvement,
)
from .alias import probability_of_improvement as probability_of_improvement
from .alias import q_expected_improvement as q_expected_improvement
from .alias import q_log_expected_improvement as q_log_expected_improvement
from .alias import q_probability_of_improvement as q_probability_of_improvement
from .alias import q_upper_confidence_bound as q_upper_confidence_bound
from .alias import upper_confidence_bound as upper_confidence_bound
from .base import Acquisition as Acquisition
